# -*- coding: utf-8 -*-
# This software is distributed under the two-clause BSD license.

import unittest

from volatildap import core

# A valid LDIF.
# Note that entries are sorted by increasing DN length, then by alphabetical
# DN order.
# Attributes are sorted in alphabetical order.
VALID_LDIF = b"""version: 1

dn: ou=groups,dc=example,dc=org
objectClass: organizationalUnit
ou: groups

dn: ou=people,dc=example,dc=org
objectClass: organizationalUnit
ou: people

dn: ou=admins,ou=people,dc=example,dc=org
objectClass: organizationalUnit
ou: admins
"""

VALID_ENTRIES = {
    'ou=people,dc=example,dc=org': {
        'ou': [b'people'],
        'objectClass': [b'organizationalUnit'],
    },
    'ou=groups,dc=example,dc=org': {
        'ou': [b'groups'],
        'objectClass': [b'organizationalUnit'],
    },
    'ou=admins,ou=people,dc=example,dc=org': {
        'ou': [b'admins'],
        'objectClass': [b'organizationalUnit'],
    },
}


class CoreTests(unittest.TestCase):
    def test_ldif_to_entries(self):
        entries = core.ldif_to_entries(VALID_LDIF)
        self.assertEqual(VALID_ENTRIES, entries)

    def test_entries_to_ldif(self):
        ldif = '\n'.join(core.entries_to_ldif(VALID_ENTRIES))
        self.assertEqual(VALID_LDIF.decode('ascii'), ldif)

    def test_loop_from_ldif(self):
        ldif = '\n'.join(core.entries_to_ldif(core.ldif_to_entries(VALID_LDIF)))
        self.assertEqual(VALID_LDIF, ldif.encode('ascii'))

    def test_loop_from_entries(self):
        entries = core.ldif_to_entries(
            '\n'.join(core.entries_to_ldif(VALID_ENTRIES)).encode('ascii')
        )
        self.assertEqual(VALID_ENTRIES, entries)



# This software is distributed under the two-clause BSD license.

"""Core components for volatildap: base types, RFC implementations, etc."""

import base64
import collections
import re


class LdapError(Exception):
    """Exceptions for volatildap"""


class PathError(LdapError):
    """Exception for missing paths"""


class TimeoutExpired(LdapError):
    def __init__(self, message, timeout=None):
        super().__init__(message)
        self.timeout = timeout


TLSConfig = collections.namedtuple('TLSConfig', ['root', 'chain', 'certificate', 'key'])


class BaseServer:
    rootdn: str
    rootpw: str
    suffix: str
    host: str
    port: int
    tls_config: TLSConfig

    def start(self):
        """Start the server if not started. Reset it otherwise."""
        raise NotImplementedError()

    def wait(self, timeout=None):
        """Wait for the server process to stop.

        The timeout is optional, and counted in seconds.

        If the server hasn't stopped before the timeout, the code will raise a
        volatildap.TimeoutExpired exception.
        """
        raise NotImplementedError()

    def reset(self):
        """Reset the directory to the contents provided at instantiation."""
        raise NotImplementedError()

    def stop(self):
        """Stop the server process."""
        raise NotImplementedError()

    def add(self, data):
        """Add items to the directory.

        Args:
            data: {dn => {attribute => [values]}}, map a distinguised name to
                  a map of attribute / values.

        Items will be inserted by increasing dn length; this ensures that an
        object is inserted after its containing entity.
        """
        raise NotImplementedError()

    def add_ldif(self, lines):
        """Add items to the directory, from a LDIF file lines.

        Args:
            lines: bytes, the LDIF content as a suite of bytes.

        Items will be inserted by increasing dn length; this ensures that an
        object is inserted after its containing entity.
        """
        self.add(ldif_to_entries(lines))

    def get(self, dn):
        """Fetch an item based on its DistinguisedName.

        The suffix may be omitted, and will be added dynamically.

        Returns:
            dict(attribute => [values]), with attribute names as strings and
            values as bytes.
        """
        raise NotImplementedError()

    def get_ldif(self, dn):
        """Fetch an item based on its DistinguisedName.

        The suffix may be omitted, and will be added dynamically.

        Returns:
            str: a LDIF file content.
        """
        entry = self.get(dn)
        lines = self._data_as_ldif({dn: entry})
        return '\n'.join(lines)

    def _normalize_dn(self, dn):
        if not dn.endswith(self.suffix):
            return '%s,%s' % (dn, self.suffix)
        else:
            return dn

    def _data_as_ldif(self, data):
        return entries_to_ldif({
            self._normalize_dn(dn): attributes
            for dn, attributes in data.items()
        })

    @property
    def uri(self):
        if self.tls_config:
            # localhost.volatildap.org is guaranteed to point to ::1 / 127.0.0.1.
            # It matches the default included TLSConfig.
            host = 'localhost.volatildap.org' if self.host == 'localhost' else self.host
            return 'ldaps://%s:%d' % (host, self.port)
        else:
            return 'ldap://%s:%d' % (self.host, self.port)


# Valid characters for a non-base64-encoded LDIF value
_BASE_LDIF_ASCII_CODES = [
    i for i in range(20, 128)
    if chr(i) not in [' ', '<', ':']
]

_BASE_LDIF = [chr(i) for i in _BASE_LDIF_ASCII_CODES]


def ldif_encode(attr, value):
    """Encode a attribute: value pair for the LDIF format.

    See RFC2849 for details.

    Rules are:
    - Text containing only chars <= 127 except control, ' ', '<', ':'
      is passed as-is
    - Other text is encoded through UTF-8 and base64-encoded
    - Binary data is simply base64-encoded

    Returns:
        A 'key: value' or 'key:: b64value' text line.
    """
    if isinstance(value, bytes):
        if all(c in _BASE_LDIF_ASCII_CODES for c in value):
            return '%s: %s' % (attr, value.decode('ascii'))
        else:
            return '%s:: %s' % (attr, base64.b64encode(value).decode('ascii'))
    elif any(c not in _BASE_LDIF for c in value):
        return '%s:: %s' % (attr, base64.b64encode(value.encode('utf-8')).decode('ascii'))
    else:
        return '%s: %s' % (attr, value)


def ldif_to_entries(ldif_lines):
    """Convert a LDIF file to a dict of dn => attributes.

    Args:
        ldif_lines: ASCII-encoded LDIF string

    Returns:
        dict(dn => dict(attribute => list(values))), where:
        - `dn` is a string;
        - `attribute` is a string;
        - `value` is bytes.

    Note: the object's DN is not included in the attributes.
    """
    entries = {}
    for entry in ldif_lines.decode('ascii').split('\n\n'):
        if not entry.strip():
            continue
        if entry.startswith('version:'):
            if re.match('^version: +1$', entry.strip()):
                continue
            else:
                raise ValueError("Invalid LDIF file - missing 'version: 1' header")

        attributes = {}
        for line in entry.split('\n'):
            if not line.strip():
                continue
            m = re.match(r'(\w+)(:?): (.*)', line.strip())
            if m is None:
                raise ValueError("Invalid line in ldif output: %r" % line)

            field, is_extended, value = m.groups()
            if is_extended:
                value = base64.b64decode(value.encode('ascii'))
            else:
                value = value.encode('ascii')
            attributes.setdefault(field, []).append(value)
        dns = attributes.pop('dn', [b''])
        assert len(dns) == 1
        entries[dns[0].decode('utf-8')] = attributes
    return entries


def entries_to_ldif(entries):
    """Convert a dict of dn => attributes to a LDIF file.

    Args:
        entries: dict(dn => dict(attribute => list(value))), where:
            - `dn` is a string;
            - `attribute` is a string;
            - `value` is bytes.

    Yields:
        str: lines of the file
    """
    yield 'version: 1'
    yield ''
    # Sort by dn length, thus adding parents first.
    for dn, attributes in sorted(entries.items(), key=lambda e: (len(e[0]), e)):
        yield ldif_encode('dn', dn)
        for attribute, values in sorted(attributes.items()):
            for value in values:
                yield ldif_encode(attribute, value)
        yield ''



# -*- coding: utf-8 -*-
# This software is distributed under the two-clause BSD license.

import os
import socket
import time
import unittest

import psutil

import volatildap


def find_available_port():
    """Find an available port.

    Simple trick: open a socket to localhost, see what port was allocated.

    Could fail in highly concurrent setups, though.
    """
    s = socket.socket()
    s.bind(('localhost', 0))
    _address, port = s.getsockname()
    s.close()
    return port


class LdapServerTestCase(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.server = None
        self.context = None

    def _launch_server(self, **kwargs):
        self.server = volatildap.LdapServer(**kwargs)
        self.server.start()
        self.context = {
            'dirname': self.server._tempdir.name,
            'pid': self.server._process.pid,
        }

    def tearDown(self):
        if self.server is not None:
            self.server.stop()
            self.assertServerStopped(self.context)
        super().tearDown()

    def assertServerStopped(self, context, max_delay=5):
        now = time.time()
        # Allow some time for proper shutdown
        while time.time() < now + max_delay and os.path.exists(context['dirname']):
            time.sleep(0.2)

        self.assertFalse(os.path.exists(context['dirname']))

        # Check that the process is no longer running.
        # We cannot rely solely on "the pid is no longer running", as it may
        # have been reused by the operating system.
        # If a process by that pid still exists, we'll check that we aren't its parent.
        try:
            stats = psutil.Process(context['pid'])
            ppid = stats.ppid()
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            # Process has died / is not in our context: all is fine.
            return

        self.assertNotEqual(ppid, os.getpid())


class ReadWriteTests(LdapServerTestCase):
    data = {
        'ou=test': {
            'objectClass': ['organizationalUnit'],
            'ou': ['test'],
        },
    }

    def test_initial_data(self):
        self._launch_server(initial_data=self.data)
        entry = self.server.get('ou=test,dc=example,dc=org')
        self.assertEqual({
            'objectClass': [b'organizationalUnit'],
            'ou': [b'test'],
        }, entry)

    def test_post_start_add(self):
        self._launch_server()
        self.server.add(self.data)
        entry = self.server.get('ou=test,dc=example,dc=org')
        self.assertEqual({
            'objectClass': [b'organizationalUnit'],
            'ou': [b'test'],
        }, entry)

    def test_get_missing_entry(self):
        self._launch_server()
        with self.assertRaises(KeyError):
            self.server.get('ou=test,dc=example,dc=org')

    def test_clear_data(self):
        self._launch_server()
        self.server.add(self.data)
        self.assertIsNotNone(self.server.get('ou=test,dc=example,dc=org'))

        self.server.start()  # Actually a restart

        # Custom data has been removed
        with self.assertRaises(KeyError):
            self.server.get('ou=test,dc=example,dc=org')

        # Core data is still there
        self.assertIsNotNone(self.server.get('dc=example,dc=org'))


class ResetTests(LdapServerTestCase):
    data = {
        'ou=test': {
            'objectClass': ['organizationalUnit'],
            'ou': ['test'],
        },
    }

    def test_cleanup(self):
        self._launch_server(initial_data=self.data)
        extra = {
            'ou=subarea,ou=test': {
                'objectClass': ['organizationalUnit'],
            },
        }
        self.server.add(extra)

        entry = self.server.get('ou=subarea,ou=test,dc=example,dc=org')
        self.server.reset()

        # Extra data should have been removed
        self.assertRaises(KeyError, self.server.get, 'ou=subarea,ou=test,dc=example,dc=org')

        # Initial data should still be here
        entry = self.server.get('ou=test,dc=example,dc=org')
        self.assertEqual({
            'objectClass': [b'organizationalUnit'],
            'ou': [b'test'],
        }, entry)


class TLSTests(LdapServerTestCase):
    def test_connection(self):
        self._launch_server(tls_config=volatildap.LOCALHOST_TLS_CONFIG)
        self.assertEqual(self.server.uri[:8], 'ldaps://')
        entry = self.server.get('dc=example,dc=org')
        self.assertEqual([b'example'], entry['dc'])


class AutoCleanupTests(LdapServerTestCase):

    def test_stop(self):
        """Deleting the LdapServer object causes its cleanup."""

        self._launch_server()


class ControlTests(LdapServerTestCase):
    def setUp(self):
        super().setUp()
        self.proxy = None

    def tearDown(self):
        if self.proxy:
            self.proxy.stop()
        super().tearDown()

    def _launch_server(self, **kwargs):
        control_port = find_available_port()
        super()._launch_server(
            control_address=('localhost', control_port),
            **kwargs,
        )
        self.proxy = volatildap.ProxyServer('http://localhost:%d/' % control_port)

    def test_launch_control(self):
        self._launch_server()
        self.assertEqual(
            self.server.uri,
            self.proxy.uri,
        )
        self.assertEqual('dc=example,dc=org', self.proxy.suffix)
        self.assertEqual(self.server.rootdn, self.proxy.rootdn)
        self.assertEqual(self.server.rootpw, self.proxy.rootpw)

    def test_tls(self):
        """The server CA should be available through the proxy."""
        self._launch_server(tls_config=volatildap.LOCALHOST_TLS_CONFIG)
        self.assertEqual(self.proxy.uri[:8], 'ldaps://')
        self.assertIsNotNone(self.proxy.tls_config)
        self.assertIsNotNone(self.proxy.tls_config.root)

    def test_get(self):
        initial = {'ou=people': {
            'objectClass': ['organizationalUnit'],
            'ou': ['people'],
        }}
        self._launch_server(
            initial_data=initial,
        )

        entry = self.proxy.get('ou=people')
        self.assertEqual(
            {
                'objectClass': [b'organizationalUnit'],
                'ou': [b'people'],
            },
            entry,
        )

    def test_add(self):
        self._launch_server()
        data = {'ou=people': {
            'objectClass': ['organizationalUnit'],
            'ou': ['people'],
        }}
        self.proxy.add(data)
        entry = self.proxy.get('ou=people')
        self.assertEqual(
            {
                'objectClass': [b'organizationalUnit'],
                'ou': [b'people'],
            },
            entry,
        )

    def test_reset(self):
        self._launch_server()
        data = {'ou=people': {
            'objectClass': ['organizationalUnit'],
            'ou': ['people'],
        }}
        self.proxy.add(data)
        # Ensure the data is visible
        self.proxy.get('ou=people')
        self.proxy.reset()

        with self.assertRaises(KeyError):
            self.proxy.get('ou=people')



# This software is distributed under the two-clause BSD license.

import http.server
import json
import subprocess
import sys
import threading
import time
from urllib.parse import urljoin

import requests

from . import core

if sys.version_info < (3, 7):
    HTTPServer = http.server.HTTPServer
else:
    HTTPServer = http.server.ThreadingHTTPServer


class ControlServer(HTTPServer):
    """The HTTP control server.

    Launched in a background thread; keeps a reference to the actual
    server.LdapServer instance.
    """
    def __init__(self, server_address, ldap_server):
        super().__init__(server_address, RequestHandler)
        self.ldap_server = ldap_server
        self._thread = None

    def start(self):
        if self._thread is not None:
            # Already started
            return
        self._thread = threading.Thread(
            target=self.serve_forever,
            name='control-server',
            daemon=True,
        )
        self._thread.start()

    def stop(self):
        if self._thread is None:
            return
        self.shutdown()
        self.server_close()
        self._thread.join()
        self._thread = None


class RequestHandler(http.server.BaseHTTPRequestHandler):
    @property
    def ldap(self):
        return self.server.ldap_server

    def _send_empty(self, status_code, message=None):
        """Send an empty HTTP response, with a specific status code."""
        self.send_response(status_code, message=message)
        self.end_headers()

    def do_POST(self):
        if self.path.strip('/') == 'control/reset':
            self.ldap.reset()
            self._send_empty(204)
        elif self.path.strip('/') == 'control/stop':
            self.ldap.stop()
            self._send_empty(204)
        elif self.path.strip('/') == 'control/start':
            self.ldap.start()
            self._send_empty(204)
        elif self.path.strip('/') == 'entry':
            self.post_entries()
        else:
            self._send_empty(404)

    def do_GET(self):
        if self.path.startswith('/entry/'):
            self.get_entry(self.path[len('/entry/'):])
        elif self.path.strip('/') == 'config':
            self.get_config()
        elif self.path.strip('/') == 'control/wait':
            self.get_wait()
        else:
            self._send_empty(404)

    def post_entries(self):
        length = int(self.headers['Content-Length'])
        data = self.rfile.read(length)
        self.ldap.add_ldif(data)
        self._send_empty(201)

    def get_config(self):
        tls_config = self.ldap.tls_config
        data = dict(
            suffix=self.ldap.suffix,
            rootdn=self.ldap.rootdn,
            rootpw=self.ldap.rootpw,
            port=self.ldap.port,
            host=self.ldap.host,
            tls_root=tls_config.root if tls_config else None
        )
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(data).encode('utf-8'))

    def get_entry(self, dn):
        try:
            entry = self.ldap.get_ldif(dn)
        except KeyError as e:
            self._send_empty(404, str(e))
        except RuntimeError as e:
            self._send_empty(500, str(e))
        else:
            self.send_response(200)
            self.send_header("Content-Type", "text/ldif")
            self.end_headers()
            self.wfile.write(entry.encode('utf-8'))

    def get_wait(self):
        try:
            self.ldap.wait(5)
        except subprocess.TimeoutExpired as e:
            self._send_empty(504, str(e))
        else:
            self._send_empty(204)


class ProxyServer(core.BaseServer):
    """A proxy to an LDAP server, based on its control API."""
    def __init__(self, url):
        self.base_url = url
        config = self._get_config()
        self.rootdn = config['rootdn']
        self.rootpw = config['rootpw']
        self.suffix = config['suffix']
        self.port = config['port']
        self.host = config['host']
        if config['tls_root']:
            self.tls_config = core.TLSConfig(
                root=config['tls_root'],
                chain=None,
                certificate=None,
                key=None,
            )
        else:
            self.tls_config = None

    def _path(self, path):
        return urljoin(self.base_url, path)

    def _get_config(self):
        response = requests.get(self._path('config/'))
        response.raise_for_status()
        return response.json()

    def reset(self):
        response = requests.post(self._path('control/reset/'))
        response.raise_for_status()

    def get(self, dn):
        response = requests.get(self._path('entry/') + dn)
        if response.status_code == 404:
            raise KeyError(dn)
        response.raise_for_status()
        entries = core.ldif_to_entries(response.content)
        assert len(entries) == 1
        return list(entries.values())[0]

    def add(self, data):
        lines = '\n'.join(core.entries_to_ldif(data))
        response = requests.post(self._path('entry/'), data=lines.encode('ascii'))
        response.raise_for_status()

    def start(self):
        response = requests.post(self._path('control/start/'))
        response.raise_for_status()

    def wait(self, timeout=None):
        since = time.time()
        while timeout is None or time.time() - since < timeout:
            response = requests.get(self._path('control/wait/'))
            if response.status_code == 504:
                continue
            else:
                response.raise_for_status()
                return
        raise core.TimeoutExpired('', timeout=timeout)

    def stop(self):
        response = requests.post(self._path('control/stop/'))
        response.raise_for_status()



# This software is distributed under the two-clause BSD license.

import argparse
import logging
import sys

from . import LOCALHOST_TLS_CONFIG
from . import core
from . import server


def launch(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--port', default='',
        help="Port to listen on; empty for a dynamic port",
    )
    parser.add_argument(
        '--host', default='localhost',
        help="Host to listen on; defaults to localhost",
    )
    parser.add_argument(
        '--suffix', default=server.DEFAULT_SUFFIX,
        help="LDAP suffix",
    )
    parser.add_argument(
        '--rootdn', default=server.DEFAULT_ROOTDN,
        help="Distinguished Name of LDAP admin user",
    )
    parser.add_argument(
        '--rootpw', default='',
        help="Password of LDAP admin user",
    )
    parser.add_argument(
        '--debug', default=server.DEFAULT_SLAPD_DEBUG, type=int,
        help="slapd debug level",
    )
    parser.add_argument(
        '--control',
        help="Start the HTTP control server on this address",
    )
    parser.add_argument(
        '--initial', type=argparse.FileType('rb'),
        help="Load initial objects from the provided LDIF file",
    )
    parser.add_argument(
        '--schemas', nargs='*', default=server.DEFAULT_SCHEMAS,
        help="Schemas to load (multi-valued)",
    )
    parser.add_argument(
        '--tls', action='store_true',
        help="Enable TLS, using a built-in stack",
    )

    args = parser.parse_args(argv)
    if args.initial:
        lines = b''.join(args.initial)
        initial = core.ldif_to_entries(lines)
    else:
        initial = {}
    if args.tls:
        tls_config = LOCALHOST_TLS_CONFIG
    else:
        tls_config = None

    if args.control:
        address, port = args.control.rsplit(':', 1)
        control_address = (address, int(port))
    else:
        control_address = ()

    instance = server.LdapServer(
        suffix=args.suffix,
        rootdn=args.rootdn,
        rootpw=args.rootpw,
        schemas=args.schemas,
        port=int(args.port) if args.port else None,
        host=args.host,
        slapd_debug=args.debug,
        initial_data=initial,
        tls_config=tls_config,
        control_address=control_address,
    )
    instance.start()
    sys.stdout.write("LDAP server listening on %s\n" % instance.uri)
    sys.stdout.write("Credentials: %s / %s\n" % (instance.rootdn, instance.rootpw))
    if control_address:
        sys.stdout.write("Control server listening on %s\n" % (instance.control.server_address,))

    try:
        instance.wait()
    finally:
        instance.stop()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    launch(sys.argv[1:])
